#line 1148 "ASTGRC.nw"
#include "ASTGRC.hpp"

namespace ASTGRC {
  
#line 208 "ASTGRC.nw"
Status Surface::visit(Pause &s) {
  Enter *en = new Enter();
  en->st=environment.grc2st[(GRCNode *)&s];
  std::cerr<<"Pause surf st="<<en->st<<std::endl;	

  *en >> context(1);
  context(0) = en;
  return Status();
}
#line 346 "ASTGRC.nw"
Status Surface::visit(IfThenElse &s) {
  Leave *lv;
  Enter *en;
  assert(s.predicate);
  lv = new Leave(); lv->st=environment.grc2st[(GRCNode*)&s];
  push_onto(context(0), lv);
  Test *t = new Test(s.predicate);
     t->st = environment.grc2st[(GRCNode*)&s];
  *t >> ( (s.else_part != 0) ? recurse(s.else_part) : context(0))
     >> ( (s.then_part != 0) ? recurse(s.then_part) : context(0));
  context(0) = t;
  en = new Enter(); en->st=environment.grc2st[(GRCNode*)&s];
  push_onto(context(0), en);
  return Status();
}
#line 418 "ASTGRC.nw"
Status Surface::visit(StatementList &s) {
  Leave *lv;
  Enter *en;
  lv=new Leave();
  lv->st=environment.grc2st[(GRCNode*)&s];
  push_onto(context(0), lv);
  for ( vector<Statement*>::reverse_iterator i = s.statements.rbegin() ;
	i != s.statements.rend() ; i++ ) {
    assert(*i);
    context(0) = synthesize(*i);
  }
  en=new Enter();
  en->st=environment.grc2st[(GRCNode*)&s];
  push_onto(context(0), en);
  return Status();
}
#line 491 "ASTGRC.nw"
Status Surface::visit(Loop &s) {
  Enter *en;
  context(0) = synthesize(s.body);
  en=new Enter(); en->st=environment.grc2st[(GRCNode*)&s];
  push_onto(context(0), en);
  return Status();
}
#line 527 "ASTGRC.nw"
Status Surface::visit(Suspend &s) {
  // to take care of immediate and counters !
  Enter *enimleaf;
  Test *tst;
  GRCNode *start;
  StartCounter *scnt;

   start = synthesize(s.body);

   Delay *d = dynamic_cast<Delay*>(s.predicate);
   if(d && d->is_immediate){
     enimleaf=new Enter();
     enimleaf->st = (STNode*) environment.grc2st[(GRCNode*)&s]->successors.back();
     *enimleaf >> context(1);
     tst = new Test(d->predicate);
     tst->st = NULL;
     *tst >> start >> enimleaf;
     start = tst;
	std::cerr<<"immediate suspend\n";
   }
   if(d && !d->is_immediate){ // a counter
	scnt = new StartCounter(d->predicate, d->count, d->counter);
	  scnt->st =  (STNode*) environment.grc2st[(GRCNode*)&s]->successors[0];
	push_onto(start, new Action(scnt));
   }

  context(0) = start;

  return Status();
}
#line 640 "ASTGRC.nw"
Status Surface::visit(Abort &s) {
 // Leave *lv;
  Enter *en;
  RecT1 *rt1;
  Nop *nop, *flowin;
  GRCNode *start;
  StartCounter *scnt;
 
//  no leaves any more !
//  lv=new Leave(); lv->st=environment.grc2st[(GRCNode*)&s];
//  push_onto(context(0), lv);

  if (s.is_weak) {
    assert(0);
  } else {

    context.push();

    rt1 = new RecT1();
    flowin = new Nop(); flowin->setflowin();  
    *flowin >> rt1;
    en = new Enter();
    en ->st = (STNode*) environment.grc2st[(GRCNode*)&s]->successors.back();

    *rt1 >> en; *en >> context(1);

     for ( int i = 0 ; i < context.size ; i++ ){
	if(i==1) continue;
	nop = new Nop(); nop->code = i;
	*nop >> rt1;
	push_onto(context(i), nop);
     }

    assert(s.body);
    start = recurse(s.body);

    context.pop();
    push_onto(start, flowin);

    for ( vector<PredicatedStatement*>::reverse_iterator i = s.cases.rbegin() ;
	  i != s.cases.rend() ; i++ ) {
      assert(*i);
      Delay *d = dynamic_cast<Delay*>((*i)->predicate);
      if (d) {
	if (d->is_immediate) {

	  // An immediate predicate: add a test an a handler

	  assert(d->predicate);
	  Test *tst = new Test(d->predicate);
	     tst->st = (STNode*)environment.grc2st[(GRCNode*)&s]->successors.back();
	  *tst >> start;
	  if ((*i)->body)
	    *tst >> recurse((*i)->body);
	  else
	    *tst >> context(0);
	  start = tst;

	} else {

	  // A counted predicate: add code that initializes the counter

	  if (d->count) {
	    assert(d->counter);
	    scnt = new StartCounter(d->predicate, d->count, d->counter);
		scnt -> st = (STNode*)environment.grc2st[(GRCNode*)&s]->successors.back();
	    push_onto(start, new Action(scnt));
	  }
	}
      }
    }

      context(0) = start;
//    en=new Enter(); en->st=environment.grc2st[(GRCNode*)&s];
//    push_onto(context(0), en);
  }

  return Status();
}
#line 877 "ASTGRC.nw"
Status Surface::visit(ParallelStatementList &s) { 
  Fork *fork = new Fork();
  Sync *sync = new Sync();
  Leave *lv;
  Enter *en;
  Terminate *t0;
  int nthr, have_t0;

  push_onto(context(0), lv=new Leave());
  lv->st=environment.grc2st[(GRCNode*)&s];
  sync->st=environment.grc2st[(GRCNode*)&s];

  GRCNode **outer = context.continuations.top();
  assert(outer);
  context.push();

  // Create a new terminate for every possible exit level
  // and link each from the sync node
  
  for ( int i = 1 ; i < context.size ; i++ )
    context(i) = new Terminate(i);

  // Synthesize each thread's surface
  have_t0 = 0;
  for ( vector<Statement*>::iterator i = s.threads.begin() ;
	i != s.threads.end() ; i++ ) {
    assert(*i);
    nthr=i-s.threads.begin();

    // this is the self looping enter
    en=new Enter(); en->st=(STNode*)environment.grc2st[(GRCNode*)&s]->successors[nthr]->successors[0];
    t0 = new Terminate(0);
    *en >> t0;

    context(0)= en;
    *fork >> recurse(*i); // it links thread to terminates, but each thread should have its own "enter"
    if(!context(0)->predecessors.empty()) {
      // add t0 to sync here, so a sync can have a t0 for each thread
      *t0 >> sync;
      have_t0 = 1;
    } 

  }

  
  // Connect each Terminate node with predecessors (i.e., that was
  // used by the threads) to the Sync and delete the rest.

  if(have_t0) *sync >> outer[0];

  for ( int i = 1 ; i < context.size ; i++ )
    if (context(i)) {
      if(!context(i)->predecessors.empty()) {
	*context(i) >> sync;
	*sync >> outer[i];
      } else {
	*sync >> 0;
	delete context(i);
      }
    }

  context.pop();
  context(0) = fork;
  push_onto(context(0), en=new Enter());
  en->st=environment.grc2st[(GRCNode*)&s];
  return Status();
}
#line 1031 "ASTGRC.nw"
Status Surface::visit(Trap &s) {
  Leave *lv;
  Enter *en;

  lv=new Leave(); lv->st=environment.grc2st[(GRCNode*)&s];
  push_onto(context(0), lv); 
  for (SymbolTable::iterator i = s.symbols->begin() ;
       i != s.symbols->end() ; i++) {
    TrapSymbol *ts = dynamic_cast<TrapSymbol*>(*i);
    assert(ts);
    context(ts->code) = context(0);
  }

  assert(s.body);
  context(0) = synthesize(s.body);
  en=new Enter(); en->st=environment.grc2st[(GRCNode*)&s];
  push_onto(context(0), en);
  return Status();
}
#line 1098 "ASTGRC.nw"
Status Surface::visit(Signal &s) {
  //  push_onto(context(0), new UndefineSignal(s));
  context(0) = synthesize(s.body);
  push_onto(context(0), new DefineSignal(s));
  return Status();
}
#line 1152 "ASTGRC.nw"
  
#line 226 "ASTGRC.nw"
Status Depth::visit(Pause &s) {
//  Leave *lv;
//  lv=new Leave();
//  lv->st=environment.grc2st[(GRCNode*)&s];
//  std::cerr<<"Pause depth st="<<lv->st<<std::endl;	
  Switch *sw;
  sw = new Switch();
  sw -> st = (STNode*) (environment.grc2st[(GRCNode*)&s])->predecessors[0];
  push_onto(context(0), sw);

  return Status();
}
#line 369 "ASTGRC.nw"
Status Depth::visit(IfThenElse &s) {
  Leave *lv;
  lv=new Leave(); lv->st=environment.grc2st[(GRCNode*)&s];
  push_onto(context(0), lv);
  Switch *sw = new Switch(); sw->st=environment.grc2st[(GRCNode*)&s];
  *sw >> ( (s.else_part != 0) ? recurse(s.else_part) : context(0))
      >> ( (s.then_part != 0) ? recurse(s.then_part) : context(0));
  context(0) = sw;
  return Status();
}
#line 437 "ASTGRC.nw"
Status Depth::visit(StatementList &s) {
  Leave *lv;
  Switch *sw;
  if (!s.statements.empty()) {
    lv=new Leave();
    lv->st=environment.grc2st[(GRCNode*)&s];
    push_onto(context(0), lv);
    sw = new Switch();
    sw->st=environment.grc2st[(GRCNode*)&s];
    environment.surface_context.push(context);
    vector<Statement*>::reverse_iterator final = s.statements.rend();
    final--;
    for ( vector<Statement*>::reverse_iterator i = s.statements.rbegin() ;
	  i != s.statements.rend() ; i++ ) {
      assert(*i);
      *sw >> synthesize(*i); // Build the depth
      // Build the surface
      if (i != final ) context(0) = environment.surface.synthesize(*i);
    }
    environment.surface_context.pop();
    context(0) = sw;
  }
  return Status();
}
#line 501 "ASTGRC.nw"
Status Depth::visit(Loop &s) {
  environment.surface_context.push(context);
  // Synthesize surface
  context(0) = environment.surface.synthesize(s.body);
  // Synthesize depth
  context(0) = synthesize(s.body);
  environment.surface_context.pop();
  return Status();
}
#line 560 "ASTGRC.nw"
Status Depth::visit(Suspend &s) {
  Switch *swimm;
  Enter *enimleaf;
  Expression *pred;
  Test *t;
  GRCNode *start;

  assert(s.predicate);
  assert(s.body);

  swimm = new Switch();
    swimm->st = environment.grc2st[(GRCNode*)&s];

  start = synthesize(s.body);

  Delay *d = dynamic_cast<Delay*>(s.predicate);
  if(d) if(d->is_immediate) pred = d->predicate;
	   else pred =  new CheckCounter(environment.boolean_type, d->counter);
     else pred = s.predicate;

  // the depth test: body is already started
  
   t = new Test(pred);
      t->st = (STNode*) environment.grc2st[(GRCNode*)&s]->successors[0];
   *t>>start>>context(1);
   *swimm >> t;

  // now the surface one
  if(d && d->is_immediate){
    enimleaf = new Enter();
    enimleaf->st = (STNode*) environment.grc2st[(GRCNode*)&s]->successors.back();
    *enimleaf >> context(1);
    t = new Test(pred);
      t->st = NULL;
    start = environment.surface.recurse(s.body);
    *t >> start >>enimleaf;
    *swimm >> t;
  }

  context(0) = swimm;
  return Status();
}
#line 722 "ASTGRC.nw"
Status Depth::visit(Abort &s) {
//  Leave *lv;

  Enter *en;
  RecT1 *rt1;
  Nop *nop, *flowin;
  GRCNode *resume, *hflow;
  Expression *pred;

//  lv=new Leave(); lv->st=environment.grc2st[(GRCNode*)&s];
//  push_onto(context(0), lv);

  if (s.is_weak) {
    assert(0);
  } else {

    context.push();

    rt1 = new RecT1();
    flowin = new Nop(); flowin->setflowin();  
    *flowin >> rt1;
    en = new Enter();
    en ->st = (STNode*) environment.grc2st[(GRCNode*)&s]->successors.back();

    *rt1 >> en; *en >> context(1);

     for ( int i = 0 ; i < context.size ; i++ ){
	if(i==1)continue;
	nop = new Nop(); nop -> code = i;
	*nop >> rt1;
	push_onto(context(i), nop);
     }

    Switch *switch_s = new Switch();
    switch_s->st= environment.grc2st[(GRCNode*)&s];

    assert(s.body);
    resume = recurse(s.body);
//    hflow = context(1);
//    push_onto(hflow, flowin);
     hflow = flowin;

    context.pop();

    for ( vector<PredicatedStatement*>::reverse_iterator i = s.cases.rbegin() ;
	  i != s.cases.rend() ; i++ ) {
        assert(*i);
        assert((*i)->predicate);

        Delay *d = dynamic_cast<Delay*>((*i)->predicate);
        if(d) if(d->is_immediate) pred = d->predicate;
		else pred = new CheckCounter(environment.boolean_type, d->counter);
	  else pred = (*i)->predicate;

	// test for body activation

        Test *tst = new Test(pred);
	  tst->st = (STNode*)environment.grc2st[(GRCNode*)&s];

        *tst >> resume >> context(1);
        resume = tst;

	// test for handler activation
	GRCNode *handler =
	  ((*i)->body) ? environment.surface.recurse((*i)->body) : context(0);
	tst = new Test(pred);	
	  tst->st = (STNode*)environment.grc2st[(GRCNode*)&s];
	*tst >> hflow >> handler;
	hflow = tst;

        // Attach the depth of this handler to the main switch

        if ((*i)->body) {
		*switch_s >> recurse((*i)->body);
        }
    }

    Nop *shnop = new Nop(); shnop->setshortcut();
    *shnop >> resume >> hflow;
    *switch_s >> shnop;
    
    context(0) = switch_s;

  }
  return Status();
}
#line 947 "ASTGRC.nw"
Status Depth::visit(ParallelStatementList &s) {
  Fork *fork = new Fork();
  Sync *sync = new Sync();
  Leave *lv;
  Enter *en;
  int nthr, have_t0;
  Terminate *t0;

  push_onto(context(0), lv = new Leave());
  lv->st=environment.grc2st[(GRCNode*)&s];
  sync->st=environment.grc2st[(GRCNode*)&s];
  sync->setdepth();

  GRCNode **outer = context.continuations.top();
  assert(outer);
  context.push();

  // Create a new terminate for every possible exit level
  // and link each from the sync node

  for ( int i = 1 ; i < context.size ; i++ )
    context(i) = new Terminate(i);

  // Synthesize each thread's surface
  have_t0=0;
  for ( vector<Statement*>::iterator i = s.threads.begin() ;
	i != s.threads.end() ; i++) {
    assert(*i);
    nthr=i-s.threads.begin();
    Switch *sw = new Switch();
    sw->st=(STNode*)(environment.grc2st[(GRCNode*)&s])->successors[nthr];
    t0=new Terminate(0);
    *fork >> sw;
    // this is the self looping enter
    en=new Enter();
    en->st=(STNode*)(environment.grc2st[(GRCNode*)&s]->successors[nthr]->successors[0]);    *sw >> en;
    *en >> t0;
    context(0)=en;
    *sw >> recurse(*i);
    if(!context(0)->predecessors.empty()) {
      // add t0 to sync here, so a sync can have a t0 for each thread
      *t0 >> sync;
      have_t0 = 1;
    }
  }

  // Connect each Terminate node with predecessors (i.e., that was
  // used by the threads) to the Sync and delete the rest.

  if(have_t0) *sync >> outer[0];
  for ( int i = 1 ; i < context.size ; i++ )
    if (context(i)) {
      if(!context(i)->predecessors.empty()) {
	*context(i) >> sync;
	*sync >> outer[i];
      } else {
	*sync >> 0;
	delete context(i);
      }
    }

  context.pop();
  context(0) = fork;
  return Status();
}
#line 1053 "ASTGRC.nw"
Status Depth::visit(Trap &s) {
  Leave *lv;
  lv=new Leave(); lv->st= environment.grc2st[(GRCNode*)&s];
  push_onto(context(0), lv);
  for (SymbolTable::iterator i = s.symbols->begin() ;
       i != s.symbols->end() ; i++) {
    TrapSymbol *ts = dynamic_cast<TrapSymbol*>(*i);
    assert(ts);
    context(ts->code) = context(0);
  }

  assert(s.body);
  context(0) = synthesize(s.body);
  return Status();
}
#line 1107 "ASTGRC.nw"
Status Depth::visit(Signal &s) {
  context(0) = synthesize(s.body);
  return Status();
}
#line 1153 "ASTGRC.nw"
  
#line 245 "ASTGRC.nw"
Status SelTree::visit(Pause &s){
  STleaf *leaf;
  STexcl *excl;
  std::cerr<<"Pause st"<<std::endl;
  leaf=new STleaf();
  excl = new STexcl(); *excl >> leaf;
  environment.grc2st[(GRCNode*)&s] = leaf;
  context(0)= excl;
  return Status();
}
#line 328 "ASTGRC.nw"
Status SelTree::visit(IfThenElse &s) {
  STexcl *ite;
  ite=new STexcl();
  environment.grc2st[(GRCNode*)&s] = ite;

  *ite >> (s.else_part ? synthesize(s.else_part) : NULL);
  *ite >> (s.then_part ? synthesize(s.then_part) : NULL);

  context(0)= ite;  
  return Status();
}
#line 398 "ASTGRC.nw"
Status  SelTree::visit(StatementList &s)
{
  STexcl *excl;

  excl=new STexcl();
  environment.grc2st[(GRCNode*)&s] = excl;

  for ( vector<Statement*>::reverse_iterator i = s.statements.rbegin() ;
	i != s.statements.rend() ; i++ ){
    assert(*i);
    *excl >> synthesize(*i);
  }

  context(0)=excl;
  std::cerr<<"seq ok"<<std::endl;
  return Status();
}


#line 480 "ASTGRC.nw"
Status SelTree::visit(Loop &s){
  STref *lp;
  lp=new STref();
  environment.grc2st[(GRCNode*)&s] = lp;
  *lp>>synthesize(s.body);
  context(0)=lp;
  return Status();
}
#line 605 "ASTGRC.nw"
Status SelTree::visit(Suspend &s)
{
  STref *sp;	
  STexcl *ex;
  Delay *d;

  ex=new STexcl(); environment.grc2st[(GRCNode*)&s] = ex;
  sp=new STref(); sp->setsuspend();
 
  *ex >> sp;
  d = dynamic_cast<Delay*>(s.predicate);
  if(d && d->is_immediate) *ex >> new STleaf();

  assert(s.body);
  *sp >> synthesize(s.body);
  context(0) = ex;
  return Status();
}
#line 811 "ASTGRC.nw"
Status SelTree::visit(Abort &s){
  STexcl *absw;
  STref *pre;

  absw=new STexcl();
  pre=new STref(); pre->setabort();

  environment.grc2st[(GRCNode*)&s] = absw; // the switch will point to the exclusive
  assert(s.body);

  std::cerr<<"Abort has "<<s.cases.size()<<" cases\n";
  for ( vector<PredicatedStatement*>::reverse_iterator i = s.cases.rbegin() ;
	  i != s.cases.rend() ; i++ ) {
      assert(*i);
      assert((*i)->predicate);

      if((*i)->body) *absw >> synthesize((*i)->body);
	else std::cerr<<"what handler is this?\n";
  }

  *pre >> synthesize(s.body);
  *absw >> pre;
  context(0) = absw;
  return Status();
}
#line 853 "ASTGRC.nw"
Status SelTree::visit(ParallelStatementList &s)
{
  STpar *par;
  STexcl *ex;
  STleaf *term;

  par=new STpar();
  environment.grc2st[(GRCNode*)&s] = par; 

  for ( vector<Statement*>::iterator i = s.threads.begin() ;
	i != s.threads.end() ; i++ ) {
    assert(*i);
    ex=new STexcl(); *par >> ex; 
    term=new STleaf(); term->setfinal(); *ex >> term;
    *ex>>synthesize(*i);
  }

  context(0)=par;
  std::cerr<<"Parallel ok"<<std::endl;
  return Status();
}
#line 1071 "ASTGRC.nw"
Status SelTree::visit(Trap &s){
  STref *tr;
  tr=new STref();
  environment.grc2st[(GRCNode*)&s] = tr;
  *tr >> synthesize(s.body);
  context(0) = tr;
  return Status();  
}
#line 1114 "ASTGRC.nw"
Status SelTree::visit(Signal &s) {
  context(0) = synthesize(s.body);
  push_onto(context(0), new STref());
  return Status();
}
#line 1154 "ASTGRC.nw"
}
