#line 1539 "Dismantle.nw"
#ifndef _DISMANTLE_HPP
#  define _DISMANTLE_HPP

#  include "AST.hpp"
#  include <assert.h>
#  include <sstream>
#  include <set>

namespace Dismantle {
  using namespace IR;
  using namespace AST;

  
#line 36 "Dismantle.nw"
Label *fresh_label();
#line 1552 "Dismantle.nw"
  
#line 80 "Dismantle.nw"
class Rewriter : public Visitor {
protected:
  Module *module;
public:
  template <class T> T* transform(T* n) {
    T* result = n ? dynamic_cast<T*>(n->welcome(*this).n) : 0;
    assert(result || !n);
    return result;
  }

  template <class T> void rewrite(T* &n) { n = transform(n); }

  StatementList& sl() { return *(new StatementList()); }

  Rewriter() : module(0) {}

  
#line 105 "Dismantle.nw"
Status visit(Modules &m) {
  for (vector<Module*>::iterator i = m.modules.begin() ;
       i != m.modules.end() ; i++ ) {
    rewrite(*i);
    assert(*i);
  }
  return &m;
}
#line 116 "Dismantle.nw"
Status visit(Module &m) {
  module = &m;
  rewrite(m.body);
  assert(m.body);
  return &m;
}
#line 125 "Dismantle.nw"
Status visit(StatementList &l) {
  for (vector<Statement*>::iterator i = l.statements.begin() ;
       i != l.statements.end() ; i++ ) {
    rewrite(*i);
    assert(*i);
  }
  return &l;
}
#line 136 "Dismantle.nw"
Status visit(ParallelStatementList &l) {
  for (vector<Statement*>::iterator i = l.threads.begin() ;
       i != l.threads.end() ; i++ ) {
    rewrite(*i);
    assert(*i);
  }
  return &l;
}
#line 147 "Dismantle.nw"
Status visit(Loop &s) {
  rewrite(s.body);
  return &s;
}

Status visit(Repeat &s) {
  rewrite(s.count);
  rewrite(s.body);
  return &s;
}

Status visit(Suspend &s) {
  rewrite(s.predicate);
  rewrite(s.body);
  return &s;
}

Status visit(Abort &s) {
  rewrite(s.body);
  for (vector<PredicatedStatement*>::iterator i = s.cases.begin() ;
       i != s.cases.end() ; i++) {
    assert(*i);
    rewrite(*i);
  }
  return &s;
}
#line 176 "Dismantle.nw"
Status visit(PredicatedStatement &s) {
  rewrite(s.predicate);
  rewrite(s.body);
  return &s;
}
#line 184 "Dismantle.nw"
Status visit(Trap &s) {
  rewrite(s.body);
  for (vector<PredicatedStatement*>::iterator i = s.handlers.begin() ;
       i != s.handlers.end() ; i++) {
    assert(*i);
    rewrite((*i)->body);
  }
  return &s;
}
#line 196 "Dismantle.nw"
Status visit(IfThenElse& n) {
  rewrite(n.predicate);
  rewrite(n.then_part);
  rewrite(n.else_part);
  return &n;
}
#line 205 "Dismantle.nw"
Status visit(Handler& s) {
  for ( vector<Catch*>::iterator i = s.catches.begin() ; 
	i != s.catches.end() ; i++ ) {
    assert(*i);
    rewrite((*i)->body);
  }
  return &s;
}
#line 216 "Dismantle.nw"
Status visit(Try& s) {
  rewrite(s.body);
  Handler &h = s;
  return visit(h);
}
#line 224 "Dismantle.nw"
Status visit(Resume& s) {
  rewrite(s.body);
  Handler &h = s;
  return visit(h);
}
#line 232 "Dismantle.nw"
Status visit(Thread& s) {
  Resume &r = s;
  return visit(r);
}
#line 239 "Dismantle.nw"
Status visit(Signal& s) {
  rewrite(s.body);
  return &s;
}
#line 246 "Dismantle.nw"
Status visit(Var& s) {
  rewrite(s.body);
  return &s;
}
#line 253 "Dismantle.nw"
Status visit(Parallel& s) {
  for ( vector<Thread*>::iterator i = s.threads.begin() ; 
        i != s.threads.end() ; i++ ) {
    assert(*i);
    rewrite(*i);
  }
  Handler &h = s;
  return visit(h);
}
#line 265 "Dismantle.nw"
Status visit(ProcedureCall& s) {
  for ( vector<Expression*>::iterator i = s.value_args.begin() ;
	i != s.value_args.end() ; i++ ) {
    assert(*i);
    rewrite(*i);
  }
  return &s;
}
#line 276 "Dismantle.nw"
Status visit(Emit& s) {
  rewrite(s.value);
  return &s;
}
#line 283 "Dismantle.nw"
Status visit(Assignment& s) {
  rewrite(s.value);
  return &s;
}
#line 290 "Dismantle.nw"
Status visit(Catch &s) {
  rewrite(s.body);
  return &s;
}
#line 297 "Dismantle.nw"
Status visit(Present& s) {
  for ( vector<PredicatedStatement*>::iterator i = s.cases.begin() ;
	i != s.cases.end() ; i++ ) {
    assert(*i);
    rewrite(*i);
  }
  if (s.default_stmt) rewrite(s.default_stmt);
  return &s;
}
#line 309 "Dismantle.nw"
Status visit(If& s) {
  for ( vector<PredicatedStatement*>::iterator i = s.cases.begin() ;
	i != s.cases.end() ; i++ ) {
    assert(*i);
    rewrite(*i);
  }
  if (s.default_stmt) rewrite(s.default_stmt);
  return &s;
}
#line 325 "Dismantle.nw"
Status visit(Nothing& n) { return &n; }
Status visit(Pause& n) { return &n; }
Status visit(Halt& n) { return &n; }
Status visit(Sustain& n) { return &n; }
Status visit(Await& n) { return &n; }
Status visit(LoopEach& n) { return &n; }
Status visit(Every& n) { return &n; }
Status visit(DoWatching& n) { return &n; }
Status visit(DoUpto& n) { return &n; }
Status visit(TaskCall& n) { return &n; }
Status visit(Exec& n) { return &n; }
Status visit(Exit& n) { return &n; }
Status visit(Run& n) { return &n; }

Status visit(Label& n) { return &n; }
Status visit(Goto& n) { return &n; }
Status visit(Break& n) { return &n; }
Status visit(Continue& n) { return &n; }
#line 348 "Dismantle.nw"
Status visit(LoadVariableExpression &e) { return &e; }
Status visit(LoadSignalExpression &e) { return &e; }
Status visit(LoadSignalValueExpression &e) { return &e; }
Status visit(LoadTrapExpression &e) { return &e; }
Status visit(LoadTrapValueExpression &e) { return &e; }
Status visit(Literal &e) { return &e; }
#line 357 "Dismantle.nw"
Status visit(UnaryOp &e) {
  rewrite(e.source);
  return &e;
}
#line 364 "Dismantle.nw"
Status visit(BinaryOp &e) {
  rewrite(e.source1);
  rewrite(e.source2);
  return &e;
}
#line 372 "Dismantle.nw"
Status visit(FunctionCall &e) {
  for ( vector<Expression*>::iterator i = e.arguments.begin() ;
	i != e.arguments.end() ; i++ ) {
    assert(*i);
    rewrite(*i);
  }
  return &e;
}
#line 383 "Dismantle.nw"
Status visit(Delay &e) {
  rewrite(e.predicate);
  rewrite(e.count);
  return &e;
}
#line 97 "Dismantle.nw"
};
#line 1553 "Dismantle.nw"
  
#line 398 "Dismantle.nw"
class Dismantler1 : public Rewriter {
public:
  
#line 427 "Dismantle.nw"
IfThenElse *dismantle_case(CaseStatement &c) {
  assert(c.cases.size() > 0);
  IfThenElse *result = 0;
  IfThenElse *lastif = 0;

  for (vector<PredicatedStatement*>::iterator i = c.cases.begin() ;
       i != c.cases.end() ; i++ ) {
    assert(*i);
    assert((*i)->predicate);
    IfThenElse *thisif = new IfThenElse((*i)->predicate);
    thisif->then_part = transform((*i)->body);
    if (result)
      lastif->else_part = thisif;
    else
      result = thisif;
    lastif = thisif;
  }
  assert(lastif);
  lastif->else_part = c.default_stmt;
  assert(result);
  return transform(result);
}

virtual Status visit(Present &s) { return dismantle_case(s); }
virtual Status visit(If &s) { return dismantle_case(s); }
#line 480 "Dismantle.nw"
Status visit(Await &a) {
  Pause *p = new Pause();
  Loop *l = new Loop(p);
  Abort *ab = new Abort(l, false);
  // Copy the predicates
  for ( vector<PredicatedStatement*>::const_iterator i = a.cases.begin();
	i != a.cases.end() ; i++ )
    ab->cases.push_back(*i);     
  return transform(ab);
}
#line 509 "Dismantle.nw"
Status visit(DoWatching &s) {
  return transform(new Abort(s.body, s.predicate, s.timeout));
}
#line 530 "Dismantle.nw"
Status visit(DoUpto &s) {
  return transform(new Abort(&(sl() << s.body << new Halt()), s.predicate, 0));
}
#line 556 "Dismantle.nw"
Status visit(LoopEach &s) {
  return transform(new Loop(new Abort(&(sl() << s.body << new Halt()),
				      s.predicate, 0)));
}
#line 595 "Dismantle.nw"
Status visit(Every &s) {
/*  Label *l = fresh_label();

  return transform(&(sl() << new Goto(l) <<
		     new Loop(new Abort(&(sl() << s.body << l << new Halt()),
				s.predicate, 0))));
*/
  Expression *pred_secondabort;
  Delay *d = dynamic_cast<Delay*>(s.predicate);

  pred_secondabort = (d && d->is_immediate) ? d->predicate : s.predicate;

  return transform(&(sl() << new Abort(new Halt(), s.predicate, 0) <<
	new Loop(new Abort(&(sl()<< s.body << new Halt()), pred_secondabort, 0))
	));
}
#line 628 "Dismantle.nw"
Status visit(Halt &s) {
  return transform(new Loop(new Pause()));
}
#line 649 "Dismantle.nw"
Status visit(Sustain &s) {
  return transform(new Loop(&(sl() <<
			      new Emit(s.signal, s.value) << new Pause())));
}
#line 660 "Dismantle.nw"
Status visit(Nothing &) {
  return transform(new StatementList());
}
#line 401 "Dismantle.nw"
};
#line 1554 "Dismantle.nw"
  
#line 674 "Dismantle.nw"
class CompletionCodes : public Visitor {
public:
  void alsoMax(AST::ASTNode *n, int &m) {
     int max = recurse(n);
     if (max > m) m = max;
  }

  int recurse(AST::ASTNode *n) {
     if (n) return n->welcome(*this).i;
     else return 0;
  }

  
#line 693 "Dismantle.nw"
Status visit(Modules &m) {
  int max = 1;
  for (vector<Module*>::iterator i = m.modules.begin() ;
       i != m.modules.end() ; i++ ) alsoMax(*i, max);
  return max;
}
#line 702 "Dismantle.nw"
Status visit(Module& s) {
  s.max_code = recurse(s.body);
  return s.max_code;
}
#line 709 "Dismantle.nw"
Status visit(Signal& s) { return recurse(s.body); }
Status visit(Var& s) { return recurse(s.body); }
Status visit(Loop &s) { return recurse(s.body); }
Status visit(Repeat &s) { return recurse(s.body); }
Status visit(Suspend &s) { return recurse(s.body); }
Status visit(PredicatedStatement &s) { return recurse(s.body); }
#line 718 "Dismantle.nw"
Status visit(StatementList &l) {
  int max = 1;
  for (vector<Statement*>::iterator i = l.statements.begin() ;
       i != l.statements.end() ; i++ ) alsoMax(*i, max);
  return max;
}
#line 727 "Dismantle.nw"
Status visit(ParallelStatementList &l) {
  int max = 1;
  for (vector<Statement*>::iterator i = l.threads.begin() ;
       i != l.threads.end() ; i++ ) alsoMax(*i, max);
  return max;
}
#line 736 "Dismantle.nw"
Status visit(IfThenElse& n) {
  int max = 1;
  alsoMax(n.then_part, max);
  alsoMax(n.else_part, max);
  return max;
}
#line 749 "Dismantle.nw"
Status visit(Pause&) { return Status(0); }
Status visit(Emit&) { return Status(0); }
Status visit(Assignment&) { return Status(0); }
Status visit(ProcedureCall&) { return Status(0); }
Status visit(TaskCall&) { return Status(0); }
Status visit(Exec&) { return Status(0); }
Status visit(Exit&) { return Status(0); }
Status visit(Run&) { return Status(0); }
Status visit(Label&) { return Status(0); }
Status visit(Goto&) { return Status(0); }
#line 767 "Dismantle.nw"
Status visit(Abort &s) {
  int max = 1;
  alsoMax(s.body, max);
  for (vector<PredicatedStatement*>::iterator i = s.cases.begin() ;
       i != s.cases.end() ; i++ ) alsoMax(*i, max);
  if (s.is_weak) {
    s.code = max + 1;
    assert(s.code >= 2);
    max += 1 + s.cases.size();
  }
  return max;
}
#line 786 "Dismantle.nw"
Status visit(Trap &s) {
  int max = 1;
  alsoMax(s.body, max);

  // FIXME: is this the right order?  Should the predicates be
  // considered before or after the code is assigned?
 
  for (vector<PredicatedStatement*>::iterator i = s.handlers.begin() ;
       i != s.handlers.end() ; i++ ) alsoMax(*i, max);

  max++; // Allocate an exit level for this trap statement

  assert(s.symbols);
  for (SymbolTable::iterator i = s.symbols->begin() ; i !=
       s.symbols->end() ; i++) {
    TrapSymbol *ts = dynamic_cast<TrapSymbol*>(*i);
    assert(ts);
    ts->code = max;
  }

  return max;
}
#line 687 "Dismantle.nw"
};
#line 1555 "Dismantle.nw"
  
#line 816 "Dismantle.nw"
class Dismantler2 : public Rewriter {
  BuiltinConstantSymbol *true_constant;
  BuiltinConstantSymbol *false_constant;
  BuiltinTypeSymbol *integer_type;
  BuiltinTypeSymbol *boolean_type;
public:

  Dismantler2() : true_constant(0), false_constant(0) {}

  VariableSymbol *new_unique_var(string, TypeSymbol*);
  void add_new_thread(Parallel *, Statement *);

  
#line 833 "Dismantle.nw"
Status visit(Module &m) {
  true_constant =
	dynamic_cast<BuiltinConstantSymbol*>(m.constants->get("true"));
  assert(true_constant);
  false_constant =
	dynamic_cast<BuiltinConstantSymbol*>(m.constants->get("false"));
  assert(false_constant);
  integer_type = dynamic_cast<BuiltinTypeSymbol*>(m.types->get("integer"));
  assert(integer_type);
  boolean_type = dynamic_cast<BuiltinTypeSymbol*>(m.types->get("boolean"));
  assert(boolean_type);

  return Rewriter::visit(m);
}
#line 862 "Dismantle.nw"
Status visit(Pause &s) {
  return transform(new Break(1));
}
#line 882 "Dismantle.nw"
Status visit(Exit &e) {
  StatementList *sl = new StatementList();
  assert(e.trap);
  assert(e.trap->code > 0);
  assert(e.trap->presence);
  assert(true_constant); // Should have been set when we visited the module
  *sl << new Assignment(e.trap->presence,
			new LoadVariableExpression(true_constant));
  if (e.value) {
    assert(e.trap->value);
    *sl << new Assignment(e.trap->value, e.value);			  
  }
  *sl << new Break(e.trap->code);
  return transform(sl);
}
#line 916 "Dismantle.nw"
Status visit(Loop &s) {
  Label *l = fresh_label(); 
  return transform(&(sl() << l << s.body << new Goto(l)));
}
#line 936 "Dismantle.nw"
Status visit(Emit &s) {
  StatementList *sl = new StatementList();
  assert(s.signal);
  assert(s.signal->presence);
  assert(true_constant); // Should have been set up when we visited the module
  *sl << new Assignment(s.signal->presence,
			new LoadVariableExpression(true_constant));
  if (s.value) {
    assert(s.signal->value);
    *sl << new Assignment(s.signal->value, s.value);   
  }
  return transform(sl);
}
#line 956 "Dismantle.nw"
Status visit(StatementList &s) {
  Rewriter::visit(s);

  StatementList *result = new StatementList();

  for (vector<Statement*>::iterator i = s.statements.begin() ;
       i != s.statements.end() ; i++) {
    StatementList *sl = dynamic_cast<StatementList*>(*i);
    if (sl)
      for (vector<Statement*>::iterator j = sl->statements.begin() ;
	   j != sl->statements.end() ; j++ ) *result << *j;
    else *result << *i;
  }

  return result;
}
#line 998 "Dismantle.nw"
Status visit(Trap &s) {
  assert(s.symbols);

  StatementList *result = new StatementList();

  int code = 0;
  int numTraps = 0;

  // Reset trap presence variables and initialize any values
  for (SymbolTable::iterator i = s.symbols->begin() ;
       i != s.symbols->end() ; i++) {
    TrapSymbol *ts = dynamic_cast<TrapSymbol*>(*i);
    assert(ts);
    code = ts->code;
    ++numTraps;

    // Reset the trap presence variable

    assert(false_constant); // Should have been set when module was visited
    *result << new Assignment(ts->presence,
	                      new LoadVariableExpression(false_constant));

    // Initialize the trap value variable if specified
    if (ts->initializer) {
      assert(ts->value); // Traps with initializers should have values
      *result << new Assignment(ts->value, ts->initializer);      
    }
  }

  // Need to check predicates if there's more than one trap
  bool need_tests = numTraps > 1;

  // Handlers run in parallel if there's more than one
  bool in_parallel = s.handlers.size() > 1;

  Statement *catchBody = 0;

  ParallelStatementList *psl;
  if (in_parallel) {
    catchBody = psl = new ParallelStatementList();
  }

  if (s.handlers.size() == 0)
    catchBody = new StatementList();
  else
    for (vector<PredicatedStatement*>::const_iterator i = s.handlers.begin() ;
         i != s.handlers.end() ; i++) {
      PredicatedStatement *handler = *i;
      assert(handler);
      Statement *body = handler->body;
      if (need_tests)
        body = new IfThenElse(handler->predicate, body, 0);
      if (in_parallel) psl->threads.push_back(body);
      else catchBody = body;  
    }
  assert(catchBody);

  Try *tryst = new Try(fresh_label(), s.body);  
  tryst->newCatch(catchBody, code, fresh_label());
  *result << tryst;

  return transform(result);
}
#line 1103 "Dismantle.nw"
Status visit(Abort &s) {
  Parallel *parallel;
  assert(boolean_type); // Module visitor should have found it
  assert(integer_type); // Module visitor should have found it

  if (!s.is_weak) parallel = 0;
  else {
    Label *c0 = fresh_label(); // Separated for determinism
    Label *cont = fresh_label();
    parallel = new Parallel(c0, cont);

    StatementList *catch1 = new StatementList();
    *catch1 << new Break(1) << new Continue();
    parallel->newCatch(catch1, 1, fresh_label());    
  }

  // For immediate tests and counter initializations
  StatementList *preamble = new StatementList();

  int code = s.code;
  assert(!s.is_weak || code >= 2); // Weak abort should have a code

  IfThenElse *firstif = 0;
  IfThenElse *previousif = 0;
  Label *elselabel = 0;
  Label *nextelselabel = 0;

  for (vector<PredicatedStatement*>::const_iterator i = s.cases.begin() ;
       i != s.cases.end() ; i++) {
    PredicatedStatement *ps = *i;
    assert(ps);
    Expression *e = ps->predicate;
    assert(e);
    Statement *handler = ps->body;

    if (s.is_weak) {
      if (!handler) handler = new StatementList();
      parallel->newCatch(handler, code, fresh_label());
      handler = new Break(code);
    }

    Delay *d = dynamic_cast<Delay*>(e);
    if (d) {
      if (d->is_immediate) {
	e = d->predicate;
	assert(e);
	StatementList *newhandler = new StatementList();
	Label *handler_label = fresh_label();
	*newhandler << handler_label;
	*preamble << new IfThenElse(e, new Goto(handler_label), 0);
	if (handler) *newhandler << handler;
	handler = newhandler;
	// Note: expression e is shared
      }

      if (d->count) {

	// A counted delay:
	// Create the count variable
	// Initialize it in the preamble
	// test and decrement in the expression

	assert(d->counter);	

	*preamble << new Assignment(d->counter, d->count);
	StatementList *newhandler = new StatementList();
	*newhandler <<
	  new Assignment(d->counter,
			 new BinaryOp(d->counter->type, "-",
				      new LoadVariableExpression(d->counter),
				      new Literal("1", d->counter->type)));
	// Tricky: to avoid a side-effect in an expression, we mimic C's &&
	// operator by adding a label in the main "else" branch.
	// Nextelselabel is that label, and either the next iterator of the
	// loop or the code that adds the final continue will instantiate
	// the label itself.

	nextelselabel = fresh_label();
	if (!handler) handler = new StatementList();
        Expression *checkneg =
	  new BinaryOp(boolean_type, "<=",
		       new LoadVariableExpression(d->counter),
		       new Literal("0", integer_type));
		       
	*newhandler << new IfThenElse(checkneg, handler,
				      new Goto(nextelselabel));
	handler = newhandler;
      } else {
	nextelselabel = 0;
      }
    }

    IfThenElse *thisif = new IfThenElse(e, handler, 0);
    if (previousif) {
      if (elselabel) {
	StatementList *sl = new StatementList();
	*sl << elselabel << thisif;
	previousif->else_part = sl;
      } else {
	previousif->else_part = thisif;
      }
    } else {
      firstif = thisif;
    }
    previousif = thisif;
    elselabel = nextelselabel;

    --code;
  }

  Label *again_label;
  if (s.is_weak) {
    again_label = fresh_label();
    *preamble << again_label << new Break(1);
  }

  // Add either continue or a goto to the last "else" clause

  assert(firstif);
  assert(previousif);
  Statement *lastbody = s.is_weak ?
    (Statement*) new Goto(again_label) : (Statement*) new Continue();
  if (elselabel) {
    StatementList *sl = new StatementList();
    *sl << elselabel << lastbody;
    lastbody = sl;
  }
  previousif->else_part = lastbody;

  // Add resume ( body ) catch 1 ( tests ... continue )
  
  assert(s.body);
  Statement *body = s.body;

  if (s.is_weak) {

    // Create the first (body) thread: add "break k" to the end
    // of the body, surround it with a resume, and add it to the parallel

    StatementList *new_body = new StatementList();
    *new_body << body << new Break(code);
    body = new_body;

    // add catch k with an empty body (k is normal termination for the body)

    parallel->newCatch(new StatementList(), code, fresh_label());

    add_new_thread(parallel, body);

    *preamble << firstif;

    add_new_thread(parallel, preamble);

    return transform(parallel);
  }

  // Strong abort

  Label *catch0 = fresh_label();
  Label *cont = fresh_label();  
  Resume *resume = new Resume(catch0, cont, body);

  StatementList *catchbody = new StatementList();
  *catchbody << new Break(1) << firstif;
  resume->newCatch(catchbody, 1, fresh_label());

  *preamble << resume;

  return transform(preamble);
}
#line 1280 "Dismantle.nw"
Status visit(LoadTrapExpression &e) {
  assert(e.trap);
  assert(e.trap->presence);
  return transform(new LoadVariableExpression(e.trap->presence));
}
#line 1288 "Dismantle.nw"
Status visit(LoadTrapValueExpression &e) {
  assert(e.trap);
  assert(e.trap->value);
  return transform(new LoadVariableExpression(e.trap->value));
}
#line 1296 "Dismantle.nw"
Status visit(LoadSignalExpression &e) {
  assert(e.signal);
  assert(e.signal->presence);
  return transform(new LoadVariableExpression(e.signal->presence));
}
#line 1304 "Dismantle.nw"
Status visit(LoadSignalValueExpression &e) {
  assert(e.signal);
  assert(e.signal->value);
  return transform(new LoadVariableExpression(e.signal->value));
}
#line 1334 "Dismantle.nw"
Status visit(ParallelStatementList &s) {
  Label *catch0 = fresh_label();
  Label *cont = fresh_label();
  Parallel *result = new Parallel(catch0, cont);

  assert(s.threads.size() > 0);

  for ( vector<Statement*>::iterator i = s.threads.begin() ;
	i != s.threads.end() ; i++ ) {
    StatementList *body = dynamic_cast<StatementList*>(*i);
    if (!body) {
      body = new StatementList();
      *body << *i;
    }
    *body << new Break(0);
    add_new_thread(result, body);
  }
  return transform(result);
}
#line 829 "Dismantle.nw"
};
#line 1556 "Dismantle.nw"
  
#line 1435 "Dismantle.nw"
class ParallelCatches : public Rewriter {
  std::set<int> *levels;
public:
  ParallelCatches() : levels(0) {}
  
#line 1447 "Dismantle.nw"
Status visit(Break &s) {
  if (levels) levels->insert(s.level);
  return &s;
}
#line 1458 "Dismantle.nw"
Status visit(Thread &s) {
  assert(levels);
  assert(levels->empty()); // Should have been initialized by our Parallel
  assert(s.body);
  rewrite(s.body);
  assert(s.catches.empty()); // Shouldn't have any catch clauses yet
  assert(s.exit_level); // Should have an exit level variable
  assert(s.exit_level->type); // with a type

  // Add a catch clause for each of the exit levels
  // catch k  { exit_level = k }
  for ( std::set<int>::iterator i = levels->begin() ;
	i != levels->end() ; i++) {
    std::ostringstream os;
    os << *i;
    s.newCatch(new Assignment(s.exit_level,
			      new Literal(os.str(),s.exit_level->type)),
	       *i, fresh_label());
  }
  return &s;
}
#line 1500 "Dismantle.nw"
Status visit(Parallel &s) {
  std::set<int> *outer = levels;

  std::set<int> parallel_levels;

  int min = 0; // ultimately, the minimum exit level considered
  for ( unsigned int k = 0 ; k < s.threads.size() ; k++ ) {
    assert(s.threads[k]);
    std::set<int> thread_levels;
    levels = &thread_levels;
    rewrite(s.threads[k]);
    assert(!thread_levels.empty()); // should have found at least one break
    std::set<int>::const_iterator i = thread_levels.begin();
    if (*i > min) min = *i;
    for ( ; i != thread_levels.end() ; i++ ) parallel_levels.insert(*i);
  }

  // Create a catch for each member of the parallel_levels set that is
  // greater than or equal to min, the maximum of the minimum levels
  // of each of the threads

  for ( std::set<int>::const_iterator i = parallel_levels.begin() ;
        i != parallel_levels.end() ; i++ )
    if (*i >= min) {
      if (outer) outer->insert(*i);
      if (*i == 1)
        s.newCatch(&(sl() <<new Break(1) <<new Continue()), 1, fresh_label());
      else
        s.newCatch(new Break(*i), *i, fresh_label());
    }

  levels = outer;
  return &s;
};
#line 1440 "Dismantle.nw"
};
#line 1557 "Dismantle.nw"
}
#endif
