#include "IR.hpp"
#include "AST.hpp"
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <set>
#include "IR.hpp"
#include "AST.hpp"
#include "GRCPrinter.hpp"
#include "GRCWalker.hpp"

namespace AST {
  
  GRCWalker::GRCWalker(TopGRC *top, STNode *stroot, Switch *top_sw, STleaf *boot, Nodemap &refmap, std::set<GRCNode *> &reachable) : top(top), stroot(stroot), top_sw(top_sw), boot(boot), refmap(refmap), reachable(reachable)
  {   
  }
  
  void GRCWalker::visit (GRCNode *n) {}

  void GRCWalker::add_children(GRCNode *n)
  {
    if(debug) std::cerr<<"try "<<refmap[n]<<'\n';
    for (vector<GRCNode*>::iterator i=n->successors.begin(); i!=n->successors.end();i++){
      if(reachable.count(*i) == 0) {
	reachable.insert(*i);
	if(debug)std::cerr<<"Node "<<refmap[*i]<<" is reachable as "<<refmap[n]<<"\'s childe\n";
	vset.insert(*i);
      }
    }
    vset.erase(n);
    alive=1;
  }

  int GRCWalker::anyleafreachable(STNode *n)
  {
    int any;

    if(!n) return 0;
    if(dynamic_cast<STleaf*>(n)) return (reachable.count(n) > 0);
    
    any=0;
    for(vector<GRCNode*>::iterator i=n->successors.begin(); i!=n->successors.end();i++)
      if((*i) && (anyleafreachable((STNode*)(*i)))) { any=1; break;}
    
    return any;
  }

  void GRCWalker::execute()
  {
    std::set<GRCNode*>::iterator it;
    debug = 1;

   
    vset.insert(top_sw); 
    reachable.insert(top_sw);
    reachable.insert(boot);
    reachable.insert(stroot);
    
    do {
      alive=0;
      for(it=vset.begin(); it!=vset.end(); it++){
	(*it)->welcome(*this);
	if(alive) break;
      }
    } while(alive);

    if(debug)  for(it=reachable.begin(); it!=reachable.end(); it++){
      std::cerr<<"Node  "<<refmap[*it]<<" is reachable\n";
    }
    

  }


  // SWITCH
  
  Status GRCWalker::visit(Switch &s)
  {
    STNode *c;
    
    int all_reach;
    
    if(debug)    std::cerr<<"try "<<refmap[&s]<<'\n';

    all_reach=1;
    for (int i = 0; i < (int)s.successors.size(); i++) {
      if (reachable.count(s.successors[i])>0) continue;
      c=(STNode*)s.st->successors[i];
      if (anyleafreachable(c)) {
	reachable.insert(s.successors[i]);
	if(debug)std::cerr<<"Node "<<refmap[s.successors[i]]<<" is reachable as ST "<<refmap[c]<<"\'s corsp\n";
	vset.insert(s.successors[i]);
	alive=1;
      } else all_reach = 0;
    }
    if(all_reach) {vset.erase(&s); alive=1;}
    return Status();
  }

  // ENTER and LEAVE

  Status GRCWalker::visit(Enter &s)
  {
    add_children(&s);
    reachable.insert(s.st); //this state is reachable
    if(debug)    std::cerr<<"STNode "<<refmap[s.st]<<" is reachable as Enter "<<refmap[&s]<<"\'s corsp\n";
    return Status(); 
  }

  Status GRCWalker::visit(Leave &s)
  {
    add_children(&s);
    return Status();
  }

  // ACTION (emit signal) and TEST (check signal)

  Status GRCWalker::visit(Action &s)
  {
    add_children(&s);
    return Status();
  }

  Status GRCWalker::visit(Test &s)
  {
    add_children(&s);
    return Status();
  }

  // TERMINATE, SYNC and FORK

  Status GRCWalker::visit(Terminate &s)
  {
    add_children(&s);
    return Status();
  }

  Status GRCWalker::visit(Sync &s)
  {
    Terminate *t;
    GRCNode *succ;
    int nothr, t0cnt, all_reach;

    nothr=s.st->successors.size();

    t0cnt=0;
    for (vector<GRCNode*>::iterator p=s.predecessors.begin(); p!=s.predecessors.end(); p++){
      t=dynamic_cast<Terminate*>(*p);
      succ=s.successors[t->code];
      if(reachable.count(*p) > 0)
	if(t->code == 0) t0cnt++; else
	  if(reachable.count(succ) == 0){
	    reachable.insert(succ);
	    vset.insert(succ);
	    alive = 1;
	  }
    }

    succ=s.successors[0];
    if(nothr == t0cnt){
      if(reachable.count(succ) == 0){
	reachable.insert(succ);
	vset.insert(succ);
	alive = 1;
      }
    }

    all_reach=1;
    for(vector<GRCNode*>::iterator c=s.successors.begin(); c!=s.successors.end(); c++){
      if(*c) if(reachable.count(*c) == 0) all_reach = 0;
    }

    if(all_reach){
      vset.erase(&s);
      alive=1;
    }

    return Status();
  }

  Status GRCWalker::visit(Fork &s)
  {
    add_children(&s);
    return Status();
  }

  Status GRCWalker::visit(Nop &s)
  {
    add_children(&s);
    return Status();
  }

  Status GRCWalker::visit(RecT1 &s)
  {
    add_children(&s);
    return Status();
  }

  Status GRCWalker::visit(DefineSignal &s)
  {
    add_children(&s);
    return Status();
  }

  Status GRCWalker::visit(UndefineSignal &s)
  {
    add_children(&s);
    return Status();
  }

}

