#include "IR.hpp"
#include "AST.hpp"
#include "GRCPrinter.hpp"
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <string.h>

using namespace std;
using namespace AST;

vector <string> sigin, sigout;

void build_sigvectors(SymbolTable *sigs)
{
  SignalSymbol *ss;

  for(SymbolTable::iterator isym = sigs->begin(); isym != sigs->end(); isym++){
    ss=dynamic_cast<SignalSymbol*>(*isym);
    if(!ss) continue;
    if(ss->direction == "output") sigout.push_back(ss->name);
    if(ss->direction == "input") sigin.push_back(ss->name);
    if(ss->direction == "inputoutput") {sigout.push_back(ss->name); sigin.push_back(ss->name);}
  }
}

void do_esterel_v5_simulator( Module *mod, char* cfilename)
{

  string mn; // module name

  mn = mod->symbol->name;

  ofstream oc(cfilename); assert(oc.good());

  oc<<"#include <stdio.h>\n";

  for(vector<string>::iterator s = sigout.begin(); s!=sigout.end(); s++){
    oc<<"int sig"<<*s<<";\n";
    oc<<"void "<<mn<<"_O_"<<*s<<"()\n{\n";
    oc<<"  sig"<<*s<<" = "<<"1;\n";
    oc<<"}\n";
  }
	

  oc<<"int main()\n{\n";
  oc<<" int active, ntick;\n";
  oc<<" ntick = 0;\n";
  // build a loop to run until program finishes
  oc<<" do {\n";
  // no input signals for now

  // init output vars to 0

  for(vector<string>::iterator s = sigout.begin(); s!=sigout.end(); s++){
    oc<<"  sig"<<*s<<" = 0;\n";
  }

  oc<<"  active = "<<mn<<"();\n";

  oc<<"printf(\"%4d \", ntick);\n";
  for(vector<string>::iterator s = sigout.begin(); s!=sigout.end(); s++){  
    oc<<"  printf(\""<<*s<<"=%d \", sig"<<*s<<");\n";
  }
  oc <<"  printf(\"\\n\");\n";

  oc<<"  ntick++;\n";
  oc<<" } while(active);\n";

  oc<<" return 0;\n}\n";
  oc.close();

}

void parse_vvp_output( char *vvpfile, char *outfile)
{
  ofstream o(outfile); assert(o.good());
  ifstream ver(vvpfile); assert(ver.good());

  char cbuf[1000], buf[1000], mbuf[10];
  int t, time, sigval, dontread;

  t=0;
  dontread=0;
  while(1){
    if(!dontread) {ver.getline(cbuf,1000); if(ver.fail()) break;}
    istringstream isc(cbuf);
    isc >> time;
    if(!isc.good()) continue; // not a time, so I am in the first lines
    time /= 1000;

    if(time == t) {strncpy(buf, cbuf, 1000); dontread=0;} else dontread=1;

    istringstream is(buf); is >> time;

    sprintf(mbuf,"%4d ",t);
    o<<mbuf;

    for(vector<string>::iterator s = sigout.begin(); s!=sigout.end(); s++){
      is >> sigval;
      o<<*s<<"="<<sigval<<" ";
    }
    t++;
    o<<'\n';
  }

  ver.close();
  o.close();

}


// output len may differ - if something goes wrong
void merge_results( char* v5filename, char *simfilename, char *outfilename)
{
  ifstream v5file(v5filename); assert(v5file.good());
  ifstream simfile(simfilename); assert(simfile.good());
  ofstream o(outfilename); assert(o.good());

  char buf5[1000], bufv[1000], mbuf[50], *name,*p;

  int t,time5, timev, x5, xv;
  string sig5, sigv;
  int gerr, lerr, err;

  int v5ok, vvpok, havesig5, havesigv;

  err=0;
  t=0;
  v5ok = vvpok = 1;
  while(1){
    lerr=0;
    if(v5ok){v5file.getline(buf5,1000); v5ok=!v5file.fail();}
    if(vvpok){simfile.getline(bufv,1000); vvpok=!simfile.fail();}

    if(!v5ok && !vvpok) break;

    if(!v5ok)  buf5[0]=0;
    if(!vvpok) bufv[0]=0;
    stringstream ss5(buf5), ssv(bufv);

    ss5>>time5; ssv>>timev;

    //    cerr<<"here: v5ok="<<v5ok<<" vvpok="<<vvpok<<" t5="<<time5<<
    //      " tv="<<timev<<" t="<<t<<"\n";

    if(v5ok) assert(t==time5);
    if(vvpok) assert(t==timev);

    sprintf(mbuf,"%4d ", t); o<<mbuf;
   
    //    cerr<<t<<"\n";

    while(1){
      havesig5 = havesigv = 0;
      if(v5ok){
	ss5>>sig5; //cerr<<"sig5="<<sig5<<"\n";
	if(!ss5.fail()) { 
	  havesig5=1;
	  strcpy(mbuf, sig5.c_str());
	  p=name=mbuf;
	  while(*p && *p!='=') p++; *p=0; p++;
	  sscanf(p,"%d", &x5);
	}
      }

      if(vvpok){
	ssv>>sigv; //cerr<<"sigv="<<sigv<<"\n";
	if(!ssv.fail()){
	  havesigv=1;
	  strcpy(mbuf, sigv.c_str());
	  p=name=mbuf;
	  while(*p && *p!='=') p++; *p=0; p++;
	  sscanf(p,"%d", &xv);
	}
      }
      if(!havesig5 && !havesigv) break;
      err = (havesig5 != havesigv || x5 != xv);

      //      if(err) break;
      if(err){gerr++; lerr++;}

      o<<name;
      if(havesig5)o<<x5; else o<<"-";
      if(havesigv)o<<xv; else o<<"-";
      o<<(err?"*":" ")<<" ";

    }
    o<<(lerr?"ERROR\n":"OK\n");
    t++;
  }
  if(gerr) o<<"ERRORS "<<gerr<<'\n'; else o<<"OK\n";

}

int main(int argc, char* argv[])
{

  char* cfilename; // for v5
  char *vvpfilename, *outfilename; // for zaphod = parse vvp
  char* v5outfilename, *simoutfilename, *mergedfilename; // for merge


  if(argc<4) {
    cerr<<"Usage: "<<argv[0]<<" v5 grcfile cfilename\n";
    cerr<<"              zaphod grcfile vvpfile outfile\n";
    cerr<<"              merge v5outfile vvpoutfile mergedfile\n";
    return 1;
  }

  string action(argv[1]);

  if(action == "merge"){
    assert(argc > 4);
    v5outfilename = argv[2];
    simoutfilename = argv[3];
    mergedfilename = argv[4];
    merge_results(v5outfilename, simoutfilename, mergedfilename);
    return 0;
  }

  ifstream grcfile(argv[2]); assert(grcfile.good());
 
  IR::XMListream r(grcfile);
  IR::Node *n;
  r >> n;
  grcfile.close();

  try{
    Modules *mods = dynamic_cast<AST::Modules*>(n);
    if (!mods) throw IR::Error("Root node is not a Modules object");
    
    for ( std::vector<AST::Module*>::iterator i = mods->modules.begin() ;
	  i != mods->modules.end() ; i++ ) {
      assert(*i);
      
      SymbolTable *sig = dynamic_cast<AST::SymbolTable*>((*i)->signals);
      assert(sig);
      
      build_sigvectors(sig);

      if(action == "v5") {
	assert(argc>3);
	cfilename = argv[3];
	do_esterel_v5_simulator(*i, cfilename);
	continue;
      }

      if(action == "zaphod") {
	assert(argc > 4);
	vvpfilename = argv[3];
	outfilename = argv[4];
	parse_vvp_output(vvpfilename, outfilename);
	continue;
      }

      cerr<<"Usage:\n"; return -1;

    }
  
  } catch (IR::Error &e) {
    std::cerr << e.s << std::endl;
    exit(-1);
  }


  return 0;
}
